# calcul des dimensions des familles de sous-espaces stables par un
# endomorphisme diagonalisable

########## guide d'utilisation

# types_stable (m,d)
# d=[d_0,...,d_{r-1}] liste des dimensions des sous-espaces propres
# m dimension du stable voulu
# renvoie la liste des listes [m_0,...,m_{r-1}] de dimensions possibles
# pour la décomposition du stable selon les sous-espaces propres

# dim_famille (f,d)
# f=[m_0,...,m_{r-1}] dimensions de la décomposition du stable selon
# les sous-espaces propres
# d=[d_0,...,d_{r-1}] dimensions des sous-espaces propres
# renvoie la dimension de la famille de sous-espaces stables
# correspondante

# max_famille_liste(m,eigdim)
# returns the largest dimension a family of m-dimensional stable subspaces
# may have for g with eigenspaces eigdim
# prints the types of stable subspaces realizing this maximum

# liste_depasses(k,sep,dlim)
# enumerates types of k-dimensional subspaces
# that move in dimension > dlim
# for the eigenspace decomposition sep

# max_famille(m,d)
# même chose que liste_max_famille mais retourne (au lieu d'imprimer)
# seulement la dimension maximale

# max_louche(d)
# dimension de la plus grande famille de stable (tous les m possibles)
# avec les espaces propres de dimension d

# liste_partitions(dim,suptaille)
# liste découpages possibles de V de dimension dim
# en somme de sous-espaces de dimensions au plus suptaille

# liste_partitions_nontriv(dim,suptaille)
# lists partitions of integer dim with chunks at most suptaille
# such there is at most one chunck of size <2

# is_smaller_0(small_list, large_list)
# tests whether small_list is smaller than large_list
# WARNING test is only partial:
# we test only if two chucks may be merged to a larger one
# which is however sufficient for our needs

# nettoie(famille)
# takes a family of eigenspace decompositions and returns the list
# of its maximal elements
# WARNING: the argument 'famille' should be sorted in the appropriate way

########## II: tests automatiques
# 

# liste_plausibles(G,n) donne la
# liste des dimensions des grassmanniennes à comparer
# (liste les cas pour lesquels il est plausible que
# le stabilisateur soit trivial)
# G = dim du groupe
# n = dim de la représentation


# teste_strate(Gr,s,d)
# imprime la liste des gags pour
# Gr = liste des [k,dim] où k = dim des sev,
#                           dim = dim de la Grassmannienne
# s = dim de la strate
# d = dim des sep pour la strate


# teste_liste_strates(Gr,Strates)
# Gr = liste des [k,dim] où k = dim des sev,
#                           dim = dim de la Grassmannienne
# Strates = liste des [s, d] où s = dim de la strate
#                               d = dim des sep pour la strate


########## code

print("\nloading toolkit for semisimple elements")

# outils de vérifications
def somme(d):
    s=0
    for x in d:
        s=s+x
    return s

# pour trier : d.sort()

# fonction intermédiaire
# pour d=[d_0,...,d_{l-1}]
# renvoie D=d_1+...+d_{l-1} et e=[d_1,...,d_{l-1}]
def dim_rec(d):
    D=0
    e=[]
    for i in range(1,len(d)):
        D=D+d[i]
        e=e+[d[i]]
    return (D,e)

# fonction intermédiaire
# pour l=[l_0,...,l_{r-1}] liste de listes
# ajoute m au début de chacune des listes l_0,...,l_{r-1}
def rajoute(m,l):
    ln=[]
    for x in l:
        ln=ln+[[m]+x]
    return ln

def types_stable(m,d):
    if len(d)==1:
        return [[m]]
    types=[]
    D,e=dim_rec(d)
    for m0 in range(max(0,m-D), min(m,d[0])+1):
        types=types+rajoute(m0,types_stable(m-m0,e))
    return types

def dim_famille(f,d):
    if len(d)==0:
        return 0
    y,x=f[:],d[:]
    y0,x0=y.pop(),x.pop()
    return (y0*(x0-y0)+dim_famille(y,x))

def max_famille_liste(m,eigdim):
    maxi=0
    liste=[]
    for f in types_stable(m,eigdim):
        dim=dim_famille(f,eigdim)
        if dim==maxi:
            liste=liste+[f]
        if dim>maxi:
            maxi=dim
            liste=[f]
    print("largest families of stable subspaces have dimension",maxi)
    print("corresponding stable subspaces decompose as follows:")
    for f in liste:
        print(f)
    return maxi

def liste_depasses(k,sep,dlim):
    depasses = []
    for f in types_stable(k,sep):
        d = dim_famille(f,sep)
        if d>dlim:
            depasses = depasses + [[f,d]]
    return depasses

def max_famille(m,d):
    maxi=0
    for f in types_stable(m,d):
        dim=dim_famille(f,d)
        if dim>maxi:
            maxi=dim
    return maxi


# max à la louche
# on somme les dimensions max de toutes les grassmanniennes

# fonction intermédiaire
# calcule le max (sur m) des dim des Gr(m,d0)
def max_gr(d0):
    maxi=0
    for m in range(d0):
        dim = m*(d0-m)
        maxi=max(dim,maxi)
    return maxi

def max_louche(d):
    louche=0
    for di in d:
        louche = louche + max_gr(di)
    return louche

def liste_partitions(dim,s):
    liste=[]
    if s==1:
        return [dim*[1]]
    for k in range(dim//s+1):
        sousliste = liste_partitions(dim-k*s,s-1)
        for i in range(len(sousliste)):
            sousliste[i]=k*[s]+sousliste[i]
        liste = sousliste + liste
    return liste

def liste_partitions_nontriv(dim,s):
    liste=[]
    if s==2:
        return [dim//2*[2]+dim%2*[1]]
    for k in range(dim//s+1):
        sousliste = liste_partitions_nontriv(dim-k*s,s-1)
        for i in range(len(sousliste)):
            sousliste[i]=k*[s]+sousliste[i]
        liste = sousliste + liste
    return liste

def is_membre(elt, liste):
    for x in liste:
        if elt==x:
            return True
    return False

def is_smaller_0(petit, gros):
    l = petit.copy()
    m = gros.copy()
    stock = []
    for x in petit:
        if is_membre(x,m):
            stock = stock + [x]
            m.remove(x)
    for x in stock:
        l.remove(x)
    stock.clear()
    i=0
    while i<len(l)-1:
        x,y=l[i],l[i+1]
        if is_membre(x+y,m):
            stock = stock + [x,y]
            m.remove(x+y)
            i=i+2
        else:
            i=i+1
    for x in stock:
        l.remove(x)
    return l==[]

def nettoie(famille):
    l=famille.copy()
    for i in range(len(l)-1,-1,-1):
        j=0
        while j<i:
            if is_smaller_0(l[i],l[j]):
                print(l[i], "smaller than", l[j])
                l.remove(l[i])
                j=i
            else:
                j=j+1
    print("maximal elements are")
    for x in l:
        print(x)
    return(l)




# implémentation tests automatiques

# liste les cas pour lesquels il est plausible que
# G = dim du groupe
# dimV = dim de la représentation
def liste_plausibles(dimG,dimV):
    l=[]
    for i in range (dimV+1):
        dim = i*(dimV-i)
        if dim >= dimG:
            l = l + [[i,dim]]
    return l

def liste_plausibles_short(dimG,dimV):
    l=[]
    for i in range (dimV//2+1):
        dim = i*(dimV-i)
        if dim >= dimG:
            l = l + [[i,dim]]
    return l

def print_plausibles(l):
    print("dimensions for which it is plausible that the general F has trivial stabilizer\n")
    for x in l:
        print("dim(F) =", x[0], ", corresp. Grass. has dim =", x[1])


# Gr = liste des [k,dim] où k = dim des sev,
#                           dim = dim de la Grassmannienne
# s = dim de la strate
# d = dim des sep pour la strate
def teste_strate(Gr,s,d):
    depasse = False 
    for gr in Gr:
        if max_famille(gr[0],d)+s >= gr[1]:
            depasse = True
            print("dépassé pour k = ", gr[0])
            liste_max_famille(gr[0], d)
    if not depasse:
        print("tutt'ok")

# Gr = liste des [k,dim] où k = dim des sev,
#                           dim = dim de la Grassmannienne
# Strates = liste des [s, d] où s = dim de la strate
#                               d = dim des sep pour la strate
def teste_liste_strates(Gr,Strates):
    for gr in Gr:
        print("\n***************","\nstable subspaces of dimension k = ", gr[0])
        print("Grassmannian has dimension", gr[1])
        depasse = False 
        for s in Strates:
            l = liste_depasses(gr[0],s[1],gr[1]-s[0]-1)
            if l!= []:
                depasse = True
                print("\nthumbsup does not hold for stratum", s)
                for x in l:
                    print("subspaces", x[0], "move in dimension", x[1])
        if not depasse:
            print("\nthumbsup holds for all tested strata")

def enumere_liste_strates(Gr,Strates):
    for gr in Gr:
        print("\n***************","\nstable subspaces of dimension k = ", gr[0])
        print("Grassmannian has dimension", gr[1])
        for s in Strates:
            print("\nexamining stratum", s)
            max = max_famille_liste(gr[0],s[1])
            if max >= gr[1]-s[0]:
                print("thumbsup DOES NOT hold for this stratum")
            else:
                print("thumbsup holds for this stratum")


##################
# definition of complete procedure
# (cosmetic presentation)

def smspl_complete (g, G, dimG, V, dimV, strates, synthetic):
    print("\n"+60*"*")
    print("genus", g)
    print("semi-simple case")
    print("")
    print ("group "+G+" : ", "dim =", dimG)
    print ("representation "+V+" : ", "dim =", dimV)
    #
    print("\n"+40*"*")
    print("list of examined strata")
    for x in strates:
        print("\neigenspace decomposition",x[1])
        print("appears in case", x[2])
        print("dim(stratum)=", x[0])
    #
    Gr = liste_plausibles(dimG,dimV)
    print("\n"+40*"*")
    print_plausibles(Gr)
    # now let's go
    print("\n"+40*"*")
    print("now let's examine all strata listed above")
    #
    if synthetic:
        teste_liste_strates(Gr,strates)
    else:
        enumere_liste_strates(Gr,strates)
    print("\n"+60*"*")

def test():
    print("smspl")

print("OK toolkit loaded")
